import pandas as pd
import numpy as np
import torch
import sqlite3
import sklearn.model_selection


def get_data_split(n: int,
                   split_ratio: float = 0.2,
                   random_seed: int = 42):
    # set seed for reproducibility
    if random_seed:
        torch.manual_seed(random_seed)
    # random indicies
    idx = torch.randperm(n)
    n_valid = int(n * split_ratio)
    idx_valid = idx[:n_valid]
    idx_train = idx[n_valid:]
    return idx_train, idx_valid


class BaseDataset(torch.utils.data.Dataset):
    def __init__(self):
        pass

    def get_validation_set(self):
        if self.idx_valid is not None:
            return self.X[self.idx_valid], self.y[self.idx_valid]
        else:
            return None, None

    def get_class_weights(self):
        cnts = torch.bincount(self.y)
        cnts = torch.maximum(cnts, torch.tensor(1))
        weights = cnts.sum() / (len(cnts) * cnts)
        return weights

    def num_classes(self):
        return len(self.labels)

    def num_features(self):
        return self.X.shape[1]

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, rowidx):
        return self.X[self.indices[rowidx]], self.y[self.indices[rowidx]]


class GermEval17(BaseDataset):
    """ ABSD-Relevance, -Sentiment, -Category
    Examples:
    ---------
    dset = GermEval17(
        preprocesser, task="Relevance", test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 task: int = 1,
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        assert task in ["Relevance", "Sentiment", "Category"]
        self.colidx = int(
            ["Relevance", "Sentiment", "Category"].index(task) + 2)

        if task == "Relevance":
            self.labels = [False, True]
        elif task == "Sentiment":
            self.labels = ['negative', 'neutral', 'positive']
        elif task == "Category":
            self.labels = [
                'Image', 'Informationen', 'Connectivity',
                'Auslastung_und_Platzangebot', 'Service_und_Kundenbetreuung',
                'Gastronomisches_Angebot', 'Allgemein', 'Design',
                'Sonstige_Unregelmässigkeiten', 'Gepäck', 'DB_App_und_Website',
                'Atmosphäre', 'Zugfahrt', 'Ticketkauf', 'nan',
                'Reisen_mit_Kindern', 'Barrierefreiheit', 'Sicherheit',
                'Toiletten', 'Komfort_und_Ausstattung']

        # read data
        split = "test" if test else "train"
        data = pd.read_csv(
            f"{datafolder}/germeval17/{split}.tsv",
            sep="\t", header=None).values
        data[:, 4] = [str(s).split(":")[0] for s in data[:, 4]]
        # bad examples to be removed
        idxbad = [i for i, x in enumerate(data[:, 1])
                  if not isinstance(x, str)]
        data = np.delete(data, idxbad, axis=0)

        # preprocess
        self.X = preprocesser(data[:, 1].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(
            [self.labels.index(row[self.colidx]) for row in data])
        # prepare data split
        if early_stopping and split == "train":
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None


class GermEval18(BaseDataset):
    """ OL18A, OL18B
    Examples:
    ---------
    dset = GermEval18(
        preprocesser, task="A", test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 task: int = 1,
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        assert task in ["A", "B", "C"]
        self.colidx = int(["A", "B", "C"].index(task) + 1)

        if task == "A":
            self.labels = ['OFFENSE', 'OTHER']
        elif task == "B":
            self.labels = ['PROFANITY', 'INSULT', 'ABUSE', 'OTHER']

        # read data
        split = "test" if test else "train"
        data = pd.read_csv(
            f"{datafolder}/germeval18/{split}.txt",
            sep="\t", header=None).values
        # preprocess
        self.X = preprocesser(data[:, 0].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(
            [self.labels.index(row[self.colidx]) for row in data])

        # prepare data split
        if early_stopping and split == "train":
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None


class GermEval19(BaseDataset):
    """ OL19A, OL19B, OL19C
    Examples:
    ---------
    dset = GermEval19(
        preprocesser, task="A", test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 task: int = 1,
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        assert task in ["A", "B", "C"]
        self.colidx = int(["A", "B", "C"].index(task) + 1)

        if task == "A":
            self.labels = ['OFFENSE', 'OTHER']
            fsuf = "12"
        elif task == "B":
            self.labels = ['PROFANITY', 'INSULT', 'ABUSE', 'OTHER']
            fsuf = "12"
        elif task == "C":
            self.labels = ['EXPLICIT', 'IMPLICIT']
            fsuf = "3"

        # read data
        split = "gold" if test else "train"
        data = pd.read_csv(
            f"{datafolder}/germeval19/{split}{fsuf}.txt",
            sep="\t", header=None).values
        # preprocess
        self.X = preprocesser(data[:, 0].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(
            [self.labels.index(row[self.colidx]) for row in data])

        # prepare data split
        if early_stopping and split == "train":
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None


class GermEval21(BaseDataset):
    """ TOXIC, ENGAGE FCLAIM
    Examples:
    ---------
    dset = GermEval21(
        preprocesser, task="TOXIC", test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 task: str = "TOXIC",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        assert task in ["TOXIC", "ENGAGE", "FCLAIM"]
        self.colidx = int(["TOXIC", "ENGAGE", "FCLAIM"].index(task) + 2)

        # read data
        split = "test" if test else "train"
        data = pd.read_csv(f"{datafolder}/germeval21/{split}.csv").values
        # preprocess
        self.X = preprocesser(data[:, 1].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor([row[self.colidx] for row in data])

        # prepare data split
        if early_stopping and split == "train":
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None

    def num_classes(self):
        return 2


class GermEval21vmwe(BaseDataset):
    """ VMWE
    Examples:
    ---------
    dset = GermEval21vmwe(
        preprocesser, test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        # self.labels = ['figuratively', 'literally', 'both', 'undecidable']
        self.labels = ['figuratively', 'literally']

        # read data
        split = "test" if test else "train"
        data = pd.read_csv(
            f"{datafolder}/germeval21vmwe/{split}.tsv",
            sep="\t", header=None).values
        # bad examples to be removed
        idxbad = [i for i, x in enumerate(data[:, 2])
                  if x not in self.labels]
        data = np.delete(data, idxbad, axis=0)

        # preprocess
        self.X = preprocesser(data[:, 3].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor([self.labels.index(row[2]) for row in data])

        # prepare data split
        if early_stopping and split == "train":
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None

    def num_classes(self):
        return 2


def merge_mio(a, b):
    s = ""
    if isinstance(a, str):
        s += a
        s += ". "
    if isinstance(b, str):
        s += b
    return s


class MillionSentiment(BaseDataset):
    """ Million Dataset
    Examples:
    ---------
    dset = MillionSentiment(
        preprocesser, test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        # self.labels = ['negative', 'neural', 'positive']
        # read data
        con = sqlite3.connect(f"{datafolder}/1mio/corpus.sqlite3")
        cur = con.cursor()
        res = cur.execute("""
        SELECT
            Posts.Headline,
            Posts.Body,
            SUM(IIF(Category='SentimentNeutral', Value, 0)),
            SUM(IIF(Category='SentimentNegative', Value, 0)),
            SUM(IIF(Category='SentimentPositive', Value, 0))
        FROM Annotations
        INNER JOIN Posts ON Posts.ID_Post=Annotations.ID_Post
        WHERE Category='SentimentNegative'
        OR Category='SentimentNeutral'
        OR Category='SentimentPositive'
        GROUP BY Annotations.ID_Post
        """)
        dat = [(merge_mio(row[0], row[1]), np.argmax(row[2:])) for row in res]
        dat = [(x, y) for x, y in dat if len(x) > 0]
        X = [str(row[0]) for row in dat]
        y = [int(row[1]) for row in dat]

        # data split
        if test:
            _, X, _, y = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)
        else:
            X, _, y, _ = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)

        # preprocess
        self.X = preprocesser(X)
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(y)

        # prepare data split
        if early_stopping and (not test):
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None

    def num_classes(self):
        return 3


class MillionBinary(BaseDataset):
    """ Million Dataset
    Tasks:
    ------
    MIO-O: 'OffTopic'
    MIO-I: 'Inappropriate'
    MIO-D: 'Discriminating'
    MIO-F: 'PossiblyFeedback'
    MIO-P: 'PersonalStories'
    MIO-A: 'ArgumentsUsed'

    Examples:
    ---------
    dset = MillionBinary(
        preprocesser, test=False, task='OffTopic',
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 task: str = "OffTopic",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        # self.labels = ['negative', 'neural', 'positive']
        # task
        t1 = ['MIO-O', 'MIO-I', 'MIO-D', 'MIO-F', 'MIO-P', 'MIO-A']
        t2 = ['OffTopic', 'Inappropriate', 'Discriminating',
              'PossiblyFeedback', 'PersonalStories', 'ArgumentsUsed']
        if task in t1:
            idx = t1.index(task)
            task = t2[idx]
        if task not in t2:
            raise Exception(f"task='{task}' does not exist")

        # read data
        con = sqlite3.connect(f"{datafolder}/1mio/corpus.sqlite3")
        cur = con.cursor()
        res = cur.execute(f"""
        SELECT
            Posts.Headline,
            Posts.Body,
            AVG(Annotations.Value) >= 0.5
        FROM Annotations
        INNER JOIN Posts ON Posts.ID_Post=Annotations.ID_Post
        WHERE Annotations.Category='{task}'
        GROUP BY Annotations.ID_Post
        """)
        dat = [(merge_mio(row[0], row[1]), row[2]) for row in res]
        dat = [(x, y) for x, y in dat if len(x) > 0]
        X = [str(row[0]) for row in dat]
        y = [int(row[1]) for row in dat]

        # data split
        if test:
            _, X, _, y = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)
        else:
            X, _, y, _ = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)

        # preprocess
        self.X = preprocesser(X)
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(y)

        # prepare data split
        if early_stopping and (not test):
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None

    def num_classes(self):
        return 2


class SBCHisSwiss(BaseDataset):
    """ SB-CH, chatmania, Swiss German detection
    Examples:
    ---------
    dset = SBCHisSwiss(
        preprocesser, test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        # read data
        df1 = pd.read_csv(f"{datafolder}/sbch/sentiment.csv")
        df2 = pd.read_csv(f"{datafolder}/sbch/chatmania.csv")
        df1['sentence_id'] = df1['sentence_id'].astype(int)
        df2['sentence_id'] = df2['sentence_id'].astype(int)
        df = df2.merge(df1, how="inner", on="sentence_id")
        y = (df["un"] == 0).astype(int).tolist()
        X = df['sentence_text'].astype(str).tolist()

        # data split
        if test:
            _, X, _, y = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)
        else:
            X, _, y, _ = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)

        # preprocess
        self.X = preprocesser(X)
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(y)

        # prepare data split
        if early_stopping and (not test):
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None

    def num_classes(self):
        return 2


class SBCHsenti(BaseDataset):
    """ SB-CH, chatmania, Sentiment Analysis, only comments detected as swiss
    Examples:
    ---------
    dset = SBCHsenti(
        preprocesser, test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        # read data
        df1 = pd.read_csv(f"{datafolder}/sbch/sentiment.csv")
        df2 = pd.read_csv(f"{datafolder}/sbch/chatmania.csv")
        df1['sentence_id'] = df1['sentence_id'].astype(int)
        df2['sentence_id'] = df2['sentence_id'].astype(int)
        df = df2.merge(df1, how="inner", on="sentence_id")
        # remove non-swiss comments
        mask_isswiss = df["un"] == 0
        df = df[mask_isswiss]
        # remove examples without sentiment
        mask_hasval = df[['neut', 'neg', 'pos']].sum(axis=1) > 0
        df = df[mask_hasval]
        # merge sentiment frequencies to class label
        y = df[['neut', 'neg', 'pos']].apply(
            lambda row: np.argmax(row), axis=1).tolist()
        X = df['sentence_text'].astype(str).tolist()

        # data split
        if test:
            _, X, _, y = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)
        else:
            X, _, y, _ = sklearn.model_selection.train_test_split(
                X, y, test_size=0.5, random_state=random_seed, stratify=y)

        # preprocess
        self.X = preprocesser(X)
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(y)

        # prepare data split
        if early_stopping and (not test):
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None

    def num_classes(self):
        return 3


class LSDC(BaseDataset):
    """ The Low Saxon Dialect Classification (LSDC) dataset

    Notes:
    ------
    - Lower Prussia (NPR) is excluded because extinct
    - Gronings (GRO) is excluded because not enough examples

    Examples:
    ---------
    dset = LSDC(
        preprocesser, test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        # excluded: 'NPR' (extinct dialect), 'GRO' (lack of data)
        self.labels = ['ACH', 'DRE', 'HAM', 'HOL', 'MAR', 'MKB', 'MON',
                       'NNI', 'OFL', 'OFR', 'OVY', 'OWL', 'SUD', 'TWE']
        # read data
        split = "test" if test else "train"
        data = pd.read_csv(
            f"{datafolder}/lsdc/{split}.tsv",
            sep="\t", header=None).values

        # bad examples to be removed
        idxbad = [i for i, x in enumerate(data[:, 0])
                  if x not in self.labels]
        data = np.delete(data, idxbad, axis=0)

        # preprocess
        self.X = preprocesser(data[:, 2].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(
            [self.labels.index(row[0]) for row in data])
        # prepare data split
        if early_stopping and split == "train":
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None


class ArchiMob(BaseDataset):
    """ ArchiMob corpus
    Examples:
    ---------
    dset = ArchiMob(
        preprocesser, test=False,
        early_stopping=True, split_ratio=0.1)
    X_valid, y_valid = dset.get_validation_set()
    n_classes = dset.num_classes()
    dgen = torch.utils.data.DataLoader(
        dset, **{'batch_size': 64, 'shuffle': True, 'num_workers': 6})
    for X, y in dgen: break
    """
    def __init__(self,
                 preprocesser,
                 datafolder: str = "datasets",
                 test: bool = False,
                 early_stopping: bool = False,
                 split_ratio: float = 0.2,
                 random_seed: int = 42):
        self.labels = ['BE', 'BS', 'ZH', 'LU']
        # read data
        if test:
            data = pd.read_csv(
                f"{datafolder}/archimob/gold.tsv",
                sep="\t", header=None).values
        else:
            data1 = pd.read_csv(
                f"{datafolder}/archimob/train.tsv",
                sep="\t", header=None).values
            data2 = pd.read_csv(
                f"{datafolder}/archimob/dev.tsv",
                sep="\t", header=None).values
            data = np.vstack([data1, data2])

        # preprocess
        self.X = preprocesser(data[:, 0].astype(str).tolist())
        if not isinstance(self.X, torch.Tensor):
            self.X = torch.tensor(self.X)
        self.y = torch.tensor(
            [self.labels.index(row[1]) for row in data])

        # prepare data split
        if early_stopping and (not test):
            self.indices, self.idx_valid = get_data_split(
                self.X.shape[0], random_seed=random_seed)
        else:
            self.indices = torch.tensor(range(self.X.shape[0]))
            self.idx_valid = None



import torch
import types
from typing import List
import sklearn.metrics
from .data import (
    GermEval17, GermEval18, GermEval19, GermEval21, GermEval21vmwe,
    MillionSentiment, MillionBinary, SBCHisSwiss, SBCHsenti, ArchiMob, LSDC)
from collections import Counter
import gc


class ClassiferModel(torch.nn.Module):
    def __init__(self,
                 input_size: int,
                 output_size: int,  # nclasses
                 bias: bool = True,
                 *args, **kwargs):
        super(ClassiferModel, self).__init__(*args, **kwargs)
        self.bias = bias
        self.final = torch.nn.Linear(
            input_size, output_size, bias=bias)
        self.soft = torch.nn.Softmax(dim=1)
        self._reset_parameters()

    def _reset_parameters(self) -> None:
        torch.manual_seed(42)
        torch.nn.init.xavier_normal_(self.final.weight, gain=1.0)
        if self.bias:
            torch.nn.init.zeros_(self.final.bias)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        x = self.final(inputs)
        return self.soft(x)


def build_model(**kwargs):
    return ClassiferModel(
        input_size=kwargs['n_features'],
        output_size=kwargs['n_classes'])


def count_and_stringify(x):
    return {str(k): str(v) for k, v in dict(Counter(x)).items()}


def evaluate(downstream_tasks: List[str],
             preprocesser: types.FunctionType,
             modelbuilder: types.FunctionType = None,
             bias: bool = True,
             datafolder: str = "./datasets",
             batch_size: int = 64,
             num_epochs: int = 20,
             balanced: bool = False,
             early_stopping: bool = False,
             split_ratio: float = 0.2,
             patience: int = 5):
    # set device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # start
    results = []
    for downstream_task in downstream_tasks:
        # load datasets
        if downstream_task == "ABSD-1":
            ds_train = GermEval17(
                preprocesser, datafolder=datafolder,
                task="Relevance", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval17(
                preprocesser, datafolder=datafolder,
                task="Relevance", test=True)
        elif downstream_task == "ABSD-2":
            ds_train = GermEval17(
                preprocesser, datafolder=datafolder,
                task="Sentiment", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval17(
                preprocesser, datafolder=datafolder,
                task="Sentiment", test=True)
        elif downstream_task == "ABSD-3":
            ds_train = GermEval17(
                preprocesser, datafolder=datafolder,
                task="Category", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval17(
                preprocesser, datafolder=datafolder,
                task="Category", test=True)

        elif downstream_task == "OL18-A":
            ds_train = GermEval18(
                preprocesser, datafolder=datafolder,
                task="A", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval18(
                preprocesser, datafolder=datafolder,
                task="A", test=True)
        elif downstream_task == "OL18-B":
            ds_train = GermEval18(
                preprocesser, datafolder=datafolder,
                task="B", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval18(
                preprocesser, datafolder=datafolder,
                task="B", test=True)

        elif downstream_task == "OL19-A":
            ds_train = GermEval19(
                preprocesser, datafolder=datafolder,
                task="A", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval19(
                preprocesser, datafolder=datafolder,
                task="A", test=True)
        elif downstream_task == "OL19-B":
            ds_train = GermEval19(
                preprocesser, datafolder=datafolder,
                task="B", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval19(
                preprocesser, datafolder=datafolder,
                task="B", test=True)
        elif downstream_task == "OL19-C":
            ds_train = GermEval19(
                preprocesser, datafolder=datafolder,
                task="C", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval19(
                preprocesser, datafolder=datafolder,
                task="C", test=True)

        elif downstream_task == "TOXIC":
            ds_train = GermEval21(
                preprocesser, datafolder=datafolder,
                task="TOXIC", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval21(
                preprocesser, datafolder=datafolder,
                task="TOXIC", test=True)
        elif downstream_task == "ENGAGE":
            ds_train = GermEval21(
                preprocesser, datafolder=datafolder,
                task="ENGAGE", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval21(
                preprocesser, datafolder=datafolder,
                task="ENGAGE", test=True)
        elif downstream_task == "FCLAIM":
            ds_train = GermEval21(
                preprocesser, datafolder=datafolder,
                task="FCLAIM", test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval21(
                preprocesser, datafolder=datafolder,
                task="FCLAIM", test=True)

        elif downstream_task == "VMWE":
            ds_train = GermEval21vmwe(
                preprocesser, datafolder=datafolder,
                test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = GermEval21vmwe(
                preprocesser, datafolder=datafolder,
                test=True)

        elif downstream_task == "MIO-S":
            ds_train = MillionSentiment(
                preprocesser, datafolder=datafolder,
                test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = MillionSentiment(
                preprocesser, datafolder=datafolder,
                test=True)
        elif downstream_task in ['MIO-O', 'MIO-I', 'MIO-D', 'MIO-F', 'MIO-P',
                                 'MIO-A']:
            ds_train = MillionBinary(
                preprocesser, datafolder=datafolder,
                test=False, task=downstream_task, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = MillionBinary(
                preprocesser, datafolder=datafolder,
                test=True, task=downstream_task)

        elif downstream_task == "SBCH-L":
            ds_train = SBCHisSwiss(
                preprocesser, datafolder=datafolder,
                test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = SBCHisSwiss(
                preprocesser, datafolder=datafolder,
                test=True)
        elif downstream_task == "SBCH-S":
            ds_train = SBCHsenti(
                preprocesser, datafolder=datafolder,
                test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = SBCHsenti(
                preprocesser, datafolder=datafolder,
                test=True)

        elif downstream_task == "ARCHI":
            ds_train = ArchiMob(
                preprocesser, datafolder=datafolder,
                test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = ArchiMob(
                preprocesser, datafolder=datafolder,
                test=True)

        elif downstream_task == "LSDC":
            ds_train = LSDC(
                preprocesser, datafolder=datafolder,
                test=False, split_ratio=split_ratio,
                early_stopping=early_stopping)
            ds_test = LSDC(
                preprocesser, datafolder=datafolder,
                test=True)
        else:
            raise Exception(
                f"Downstream task '{downstream_task}' not available.")

        # data loader
        X_valid, y_valid = ds_train.get_validation_set()
        n_classes = ds_train.num_classes()
        n_features = ds_train.num_features()
        dgen = torch.utils.data.DataLoader(
            ds_train, batch_size=batch_size, shuffle=False)

        # init new model
        if not isinstance(modelbuilder, types.FunctionType):
            modelbuilder = build_model

        model = modelbuilder(
            n_features=n_features,
            n_classes=n_classes,
            bias=bias
        ).to(device)

        # loss function
        if balanced:
            class_weights = ds_train.get_class_weights()
        else:
            class_weights = torch.ones(n_classes) / n_classes

        loss_fn = torch.nn.CrossEntropyLoss(
            weight=class_weights,
            reduction='mean'
        ).to(device)

        # early stopping
        if y_valid is not None:
            with torch.no_grad():
                X_valid, y_valid = X_valid.to(device), y_valid.to(device)
                valid_loss = loss_fn(model(X_valid), y_valid).item()
                wait = 0

        # optimization settings
        optimizer = torch.optim.Adam(
            model.parameters(), lr=3e-4, betas=(.9, .999),
            eps=1e-7, amsgrad=True)

        for epoch in range(num_epochs):
            # train
            # epoch_loss = 0.
            for X_train, y_train in dgen:
                X_train, y_train = X_train.to(device), y_train.to(device)
                torch.cuda.empty_cache()  # del previous batch from GPU RAM
                # train it
                optimizer.zero_grad()
                y_pred = model(X_train)
                loss = loss_fn(y_pred, y_train)
                loss.backward()
                optimizer.step()
                # epoch_loss += loss.item()
            # early stopping
            if y_valid is not None:
                with torch.no_grad():
                    tmp = loss_fn(model(X_valid), y_valid).item()
                    if tmp >= valid_loss:
                        wait += 1
                    else:
                        valid_loss = tmp
                        wait = 0
                    if wait > patience:
                        break

        # force memory flush
        del optimizer, X_valid, y_valid, dgen, X_train, y_train
        gc.collect()
        torch.cuda.empty_cache()

        # inference on test dataset
        dgen_test = torch.utils.data.DataLoader(
            ds_test, batch_size=batch_size, shuffle=False)
        y_test = []
        y_pred = []
        for X, y in dgen_test:
            y_test.append(y.to('cpu'))
            y_pred.append(torch.argmax(model(X.to(device)), dim=1).to('cpu'))
            torch.cuda.empty_cache()  # del previous batch from GPU RAM

        y_test = torch.cat(y_test).detach().numpy()
        y_pred = torch.cat(y_pred).detach().numpy()

        res_test = {
            "num": len(y_pred),
            "acc": float(sklearn.metrics.accuracy_score(y_test, y_pred)),
            "acc-balanced": float(sklearn.metrics.balanced_accuracy_score(
                y_test, y_pred)),
            "f1": float(sklearn.metrics.f1_score(
                y_test, y_pred, average='micro')),
            "f1-balanced": float(sklearn.metrics.f1_score(
                y_test, y_pred, average='macro')),
            "distr-pred": count_and_stringify(y_pred),
            "distr-test": count_and_stringify(y_test),
        }
        del y_pred, y_test, dgen_test, X, y
        gc.collect()
        torch.cuda.empty_cache()

        # inference on training datatsets
        dgen_train = torch.utils.data.DataLoader(
            ds_train, batch_size=batch_size, shuffle=False)
        y_train = []
        y_pred = []
        for X, y in dgen_train:
            y_train.append(y.to('cpu'))
            y_pred.append(torch.argmax(model(X.to(device)), dim=1).to('cpu'))
            torch.cuda.empty_cache()  # del previous batch from GPU RAM

        y_train = torch.cat(y_train).detach().numpy()
        y_pred = torch.cat(y_pred).detach().numpy()

        res_train = {
            "num": len(y_pred),
            "acc": float(sklearn.metrics.accuracy_score(y_train, y_pred)),
            "acc-balanced": float(sklearn.metrics.balanced_accuracy_score(
                y_train, y_pred)),
            "f1": float(sklearn.metrics.f1_score(
                y_train, y_pred, average='micro')),
            "f1-balanced": float(sklearn.metrics.f1_score(
                y_train, y_pred, average='macro')),
            "distr-pred": count_and_stringify(y_pred),
            "distr-train": count_and_stringify(y_train),
        }
        del y_pred, y_train, dgen_train, X, y
        gc.collect()
        torch.cuda.empty_cache()

        # save results
        results.append({
            "task": str(downstream_task),
            "epochs": int(epoch + 1),
            "test": res_test,
            "train": res_train
        })

        # force memory flush
        del model
        gc.collect()
        torch.cuda.empty_cache()

    # done
    return results
